#ifndef INCLUDED_ABSTRACTPARSETREE_H
#define INCLUDED_ABSTRACTPARSETREE_H

typedef struct tree_t tree_t, *tree_p;
typedef struct list_t list_t, *list_p;

class AbstractParseTree
{
public:
	AbstractParseTree() : _tree(0) {}
	AbstractParseTree(const AbstractParseTree& lhs);
	AbstractParseTree& operator=(const AbstractParseTree& lhs);
	~AbstractParseTree() { release(); }

	class iterator
	{
		friend class AbstractParseTree;
	public:
		iterator(const AbstractParseTree& tree);
		bool more();
		void next();
		
	private:
		list_p _list;
	};
	
	friend class iterator;

	AbstractParseTree(const iterator& lhs);
	
	bool isEmpty() const { return _tree == 0; }
	
	void clear();
	void attach(AbstractParseTree& tree);
	
	void createIdent( const Ident ident );
	void createStringAtom( const char *str );
	void createDoubleAtom( double value );
	void createCharAtom( char value );
	void createIntAtom( int value );
	void createList();
	void createTree( const Ident name );
	void setTreeName( const Ident name );
	void insertChild( const AbstractParseTree& child );
	void appendChild( const AbstractParseTree& child );
	void dropLastChild();
	void createOpenContext();
	void createCloseContext();
	
	bool isIdent() const;
	bool isIdent( const Ident ident ) const;
	bool equalIdent( const Ident ident ) const;
	Ident identName() const;
	bool isString() const;
	bool isString( const char *str ) const;
	const char* stringValue() const;
	bool isInt() const;
	int intValue() const;
	bool isDouble() const;
	double doubleValue() const;
	bool isChar() const;
	char charValue() const;
	bool isList() const;
	bool isTree( Ident name ) const;
	bool equalTree( Ident name ) const;
	const char* type() const;
	int nrParts() const;
	AbstractParseTree part(int n) const;
	
	void setLineColumn(int line, int column);
	int line() const;
	int column() const;

	void print(FILE* f, bool compact);
	
private:
	void release();
	void create();

	tree_p _tree;
};

const char* get_tt_opencontext();
const char* get_tt_closecontext();

#endif // INCLUDED_ABSTRACTPARSETREE_H
