#include "StdAfx.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "Ident.h"
#include "AbstractParseTree.h"
#include "TextFilePos.h"
#include "Scanner.h"
#include "AbstractParser.h"


Ident id_nt_def = "nt_def";
Ident id_rule = "rule";
Ident id_opt = "opt";
Ident id_seq = "seq";
Ident id_list = "list";
Ident id_chain = "chain";
Ident id_ident = "ident";
Ident id_identalone = "identalone";
Ident id_identdef = "identdef";
Ident id_identdefadd = "identdefadd";
Ident id_identuse = "identuse";
Ident id_identfield = "identfield";
Ident id_opencontext = "opencontext";
Ident id_closecontext = "closecontext";
Ident id_eof = "eof";

#define DEBUG_NL if (_debug_parse) printf("\n")
#define DEBUG_PT(X) if (_debug_parse) X.print(stdout, true)
#define DEBUG_PO(X) if (_debug_parse) print_or(stdout, X)
#define DEBUG_PR(X) if (_debug_parse) print_rule(stdout, X)
#define DEBUG_(X)  if (_debug_parse) printf(X)
#define DEBUG_P1(X,A) if (_debug_parse) printf(X,A)

#define RK_NT              0
#define RK_LIT             1
#define RK_OR_RULE         2
#define RK_T_EOF           3
#define RK_TERM			   4
#define RK_IDENT           5
#define RK_T_OPENCONTEXT   6
#define RK_T_CLOSECONTEXT  7

#define IK_IDENTDEF      1
#define IK_IDENTDEFADD   2
#define IK_IDENTUSE      3
#define IK_IDENTFIELD    4


class ParsedValue
{   
public:
	ParsedValue* prev;
    AbstractParseTree last;
};


enum EnumSuccess { s_unknown, s_fail, s_success } ;

class ParseSolution
{	
public:
	ParseSolution* next;
	Ident nt;
	enum EnumSuccess success;
	AbstractParseTree result;
	TextFilePos sp;
};


AbstractParser::AbstractParser()
{
	_debug_nt = true;
	_debug_parse = true;
	_debug_scan = true;
}

void AbstractParser::loadGrammer(const AbstractParseTree& root)
{
	_all_nt = 0;
	
    for (AbstractParseTree::iterator rules(root); rules.more(); rules.next())
    {
    	AbstractParseTree rule(rules);
        if (rule.isTree(id_nt_def))
        {   const char *nt_name = rule.part(1).stringValue();
            GrammarNonTerminal* nt = find_nt(nt_name);
            GrammarOrRule* *p_first = &nt->first;
            GrammarOrRule* *p_recursive = &nt->recursive;

            while (*p_first != 0)
                p_first = &(*p_first)->next;
            while (*p_recursive != 0)
                p_recursive = &(*p_recursive)->next;

            for (AbstractParseTree::iterator or_rule(rule.part(2)); or_rule.more(); or_rule.next())
            {   AbstractParseTree rule(or_rule);
                if (rule.isTree(id_rule))
                {   GrammarOrRule* new_or_rule = new GrammarOrRule;

                    if (rule.nrParts() == 0)
                    {   new_or_rule->rule = 0;
                        new_or_rule->next = 0;
                        *p_first = new_or_rule;
                        p_first = &new_or_rule->next;
                    }
                    else
                    {   AbstractParseTree::iterator parts(rule.part(1));
						if (rule.part(2).isIdent())
							new_or_rule->tree_name = rule.part(2).stringValue();
                        new_or_rule->next = 0;
                        if (parts.more() && AbstractParseTree(parts).isIdent(nt_name))
                        {
                        	parts.next();
                            new_or_rule->rule = make_rule(parts);
                            *p_recursive = new_or_rule;
                            p_recursive = &new_or_rule->next;
                        }
                        else
                        {
                            new_or_rule->rule = make_rule(parts);
                            *p_first = new_or_rule;
                            p_first = &new_or_rule->next;
                        }
                    }                     
                }
            }
        }
    }
	if(false)
    {	GrammarNonTerminal* nt;
    
    	for (nt = _all_nt; nt != 0; nt = nt->next)
    		if (nt->first == 0 && nt->recursive == 0)
    		    printf("Non-terminal '%s' has no rule.\n", nt->name.val());
    		else
    		{
    			printf("Non-terminal '%s'\n", nt->name.val());
    			
    			for (GrammarOrRule* o_r = nt->first; o_r != 0; o_r = o_r->next)
    			{
    				printf(" First: ");
    				print_rule(stdout, o_r->rule);
    				printf("\n");
    			}
    			for (GrammarOrRule* o_r2 = nt->recursive; o_r2 != 0; o_r2 = o_r2->next)
    			{
    				printf(" Recursive: ");
    				print_rule(stdout, o_r2->rule);
    				printf("\n");
    			}
    		}
    }		    
}

GrammarRule* AbstractParser::make_rule(AbstractParseTree::iterator rule)
{   GrammarRule* result;
 
    if (!rule.more())
        return 0;

    result = new GrammarRule;
    result->optional = false;
    result->sequential = false;
    result->chain_symbol = 0;
    result->last_fail_pos = (long)-1;

    AbstractParseTree elem = rule;
    /* Is the first element optional? */
    if (elem.equalTree(id_opt))
    {   result->optional = true;
        elem = elem.part(1);
    }

    /* Is it a sequential, chain or list? */
    if (elem.equalTree(id_seq))
    {   result->sequential = true;
        elem = elem.part(1);
    }
    else if (elem.equalTree(id_chain))
    {   result->sequential = true;
        result->chain_symbol = elem.part(2).stringValue();
        elem = elem.part(1);
    }
    else if (elem.equalTree(id_list))
    {   result->sequential = true;
        result->chain_symbol = ",";
        elem = elem.part(1);
    }

    if (elem.isIdent())
    {
       	Ident name = elem.identName();
       	
        if (name == id_eof)
            result->kind = RK_T_EOF;
        else
        {
        	if (_scanner->isTerminal(name))
        	{
        		result->kind = RK_TERM;
        		result->text.terminal = new GrammarTerminal(name);
        	}
        	else
        	{
	            result->kind = RK_NT;
	            result->text.non_terminal = find_nt(elem.stringValue());
	            if (result->text.non_terminal == 0)
	            {
	                printf("Undefined non-terminal: '%s'\n", elem.stringValue());
	                exit(1);
	            }
	        }
        }
    }
    else if (elem.isTree(id_identalone))
    {
        result->kind = RK_TERM;
        result->text.terminal = new GrammarTerminal(id_ident);
    }
    else if (elem.isTree(id_identdef))
    {
        result->kind = RK_IDENT;
        result->text.ident = new GrammarIdent;
        result->text.ident->kind = IK_IDENTDEF;
        result->text.ident->ident_class = elem.part(2).stringValue();
        result->text.ident->terminal = new GrammarTerminal(id_ident);
    }
    else if (elem.isTree(id_identdefadd))
    {
        result->kind = RK_IDENT;
        result->text.ident = new GrammarIdent;
        result->text.ident->kind = IK_IDENTDEFADD;
        result->text.ident->ident_class = elem.part(2).stringValue();
        result->text.ident->terminal = new GrammarTerminal(id_ident);
    }
    else if (elem.isTree(id_identuse))
    {
        result->kind = RK_IDENT;
        result->text.ident = new GrammarIdent;
        result->text.ident->kind = IK_IDENTUSE;
        result->text.ident->ident_class = elem.part(2).stringValue();
        result->text.ident->terminal = new GrammarTerminal(id_ident);
    }
    else if (elem.isTree(id_identfield))
    {
        result->kind = RK_IDENT;
        result->text.ident = new GrammarIdent;
        result->text.ident->kind = IK_IDENTFIELD;
        result->text.ident->ident_class = elem.part(2).stringValue();
        result->text.ident->terminal = new GrammarTerminal(id_ident);
    }
    else if (elem.isTree(id_opencontext))
    {
    	result->kind = RK_T_OPENCONTEXT;
    }
    else if (elem.isTree(id_closecontext))
    {
		result->kind = RK_T_CLOSECONTEXT;
	}
    else if (elem.isString())
    {
        result->kind = RK_LIT;
        result->text.str_value = elem.stringValue();
		_scanner->addKeyword(elem.stringValue());
    }
    else if (elem.isList())
    {
        result->kind = RK_OR_RULE;
        result->text.or_rule = make_or_rule(AbstractParseTree::iterator(elem));
    }
    rule.next();
    result->next = make_rule(rule);

    return result;
}

GrammarOrRule* AbstractParser::make_or_rule(AbstractParseTree::iterator or_rule)
{
    if (!or_rule.more())
        return 0;

    {   AbstractParseTree rule = or_rule;
        if (rule.isTree(id_rule))
        {   GrammarOrRule* result = new GrammarOrRule;

            if (rule.nrParts() == 0)
            {
                result->rule = 0;
            }
            else
            {   result->rule = make_rule(AbstractParseTree::iterator(rule.part(1)));
				if (rule.part(2).isIdent())
					result->tree_name = rule.part(2).identName();
            }
            or_rule.next();
            result->next = make_or_rule(or_rule);
            return result;
        }
    }
 
 	or_rule.next();
    return make_or_rule(or_rule);
}

GrammarNonTerminal* AbstractParser::find_nt(Ident name)
{
   GrammarNonTerminal* *p_nt = &_all_nt;

   while (*p_nt != 0 && (*p_nt)->name != name)
        p_nt = &((*p_nt)->next);

   if (*p_nt == 0)
   {   *p_nt = new GrammarNonTerminal;
       (*p_nt)->name = name;
       (*p_nt)->first = 0;
       (*p_nt)->recursive = 0;
       (*p_nt)->next = 0;
   }
   return *p_nt;
}

void AbstractParser::print_rule(FILE *f, GrammarRule* rule)
{   
    bool optional,
         sequential;
    const char *chain_sym;

    if (rule == 0)
        return;

    optional = rule->optional,
    sequential = rule->sequential;
    chain_sym = rule->chain_symbol;

    switch(rule->kind)
    {   case RK_T_EOF:
            fprintf(f, "<eof> ");
            break;
        case RK_TERM:
        	fprintf(f, "<%s> ", rule->text.terminal->name.val());
        	break;
        case RK_IDENT:
            fprintf(f, "<%s> ", rule->text.ident->terminal->name.val());
            break;
        case RK_NT:
            fprintf(f, "%s ", rule->text.non_terminal->name.val());
            break;
        case RK_LIT:
            fprintf(f, "\"%s\" ", rule->text.str_value);
            break;
        case RK_OR_RULE:
            fprintf(f, "(");
            print_or(f, rule->text.or_rule);
            fprintf(f, ")");
            break;
    }

    if (sequential)
    {   if (chain_sym == 0)
            fprintf(f, "SEQ ");
        else if (strcmp( chain_sym, ","))
            fprintf(f, "LIST ");
        else
            fprintf(f, "CHAIN %s ", chain_sym);
    }
    if (optional)
    {   fprintf(f, "OPT ");
    }
    print_rule(f, rule->next);
}

void AbstractParser::print_or(FILE *f, GrammarOrRule* or_rule)
{
    bool first = true;

    for (; or_rule; or_rule = or_rule->next)
    {   
        if (!first)
            fprintf(f, "|");
        first = false;
        print_rule(f, or_rule->rule);
        if (!or_rule->tree_name.empty())
            fprintf(f, "[%s]", or_rule->tree_name.val());
    }
}

void AbstractParser::expected_string(TextFilePos& pos, const char *s, bool is_keyword)
{
    if (pos < _f_file_pos)
        return;
    if (pos > _f_file_pos)
    {   _nr_exp_syms = 0;
        _f_file_pos = pos;
    }

    if (_nr_exp_syms >= 200)
        return;

    _expect[_nr_exp_syms].sym = s;
    _expect[_nr_exp_syms].in_nt = _current_nt;
    _expect[_nr_exp_syms].rule = _current_rule;
    _expect[_nr_exp_syms].is_keyword = is_keyword;
    _nr_exp_syms++;
}


void AbstractParser::getExpectedPos(long &line, long &column)
{
	line = _f_file_pos.line();
	column = _f_file_pos.column();
}

AbstractParser::ExpectAccessor::ExpectAccessor(AbstractParser* abstractParser)
  : _abstractParser(abstractParser), _i(-1)
{
	next();
}
bool AbstractParser::ExpectAccessor::more()
{
	return _i < _abstractParser->_nr_exp_syms;
}
void AbstractParser::ExpectAccessor::next()
{
	for (_i++ ;_i < _abstractParser->_nr_exp_syms; _i++)
	{
		bool unique = true;
        for (int j = 0; unique && j < _i; j++)
            if (_abstractParser->_expect[_i].rule == _abstractParser->_expect[j].rule)
                unique = false;
        if (unique)
			return;
	}
}
const char* AbstractParser::ExpectAccessor::symbol()
{
	return _abstractParser->_expect[_i].sym;
}


void AbstractParser::printExpected(FILE *f)
{   
    fprintf(f, "%d.%d Expected:\n", _f_file_pos.line(), _f_file_pos.column());
    for (int i = 0; i < _nr_exp_syms; i++)
    {   bool unique = true;
        int j;
        for (j = 0; unique && j < i; j++)
            if (_expect[i].rule == _expect[j].rule)
                unique = false;
        if (unique)
        {
            fprintf(f, "    ");
            if (_expect[i].rule)
                print_rule(stdout, _expect[i].rule);
            else
                printf("%s ", _expect[i].sym);
            fprintf(f, ":");
            if (!_expect[i].in_nt.empty())
                fprintf(f, " in %s", _expect[i].in_nt.val());
            if (_expect[i].is_keyword != 0)
                fprintf(f, " (is a keyword)");
            fprintf(f, "\n");
        }
    }
}

void AbstractParser::traceExpected()
{   
    TRACE2("%d.%d Expected:\n", _f_file_pos.line(), _f_file_pos.column());
    for (int i = 0; i < _nr_exp_syms; i++)
    {   bool unique = true;
        int j;
        for (j = 0; unique && j < i; j++)
            if (_expect[i].rule == _expect[j].rule)
                unique = false;
        if (unique)
        {
            TRACE("    ");
            //if (_expect[i].rule)
              //  TRACE("rule "); //, _expect[i].rule.val());
            //else
                TRACE1("%s ", _expect[i].sym);
            TRACE(":");
            if (!_expect[i].in_nt.empty())
                TRACE1(" in %s", _expect[i].in_nt.val());
            if (_expect[i].is_keyword != 0)
                TRACE(" (is a keyword)");
            TRACE("\n");
        }
    }
}


