#ifndef _INCLUDED_ABSTRACTPARSER_H
#define _INCLUDED_ABSTRACTPARSER_H

#include "ParserGrammar.h"

class AbstractScanner;

class AbstractParser
{
public:
	AbstractParser();
	
	void setScanner(AbstractScanner* scanner) { _scanner = scanner; }
	void loadGrammer(const AbstractParseTree& root);
	
	void setDebugLevel(bool debug_nt, bool debug_parse, bool debug_scan) { _debug_nt = debug_nt; _debug_parse = debug_parse; _debug_scan = debug_scan; }

	virtual bool parse(const char* str, Ident root_id, AbstractParseTree& result) = 0;

	void getExpectedPos(long &line, long &pos);
	class ExpectAccessor
	{
	public:
		ExpectAccessor(AbstractParser* abstractParser);
		bool more();
		void next();
		const char* symbol();
	private:
		AbstractParser* _abstractParser;
		int _i;
	};
	friend class ExpectAccessor;

	void printExpected(FILE *f);
	void traceExpected();
	
protected:
	GrammarNonTerminal* find_nt(Ident name);
	GrammarRule* make_rule(AbstractParseTree::iterator rule);
	GrammarOrRule* make_or_rule(AbstractParseTree::iterator or_rule);
	void print_rule(FILE *f, GrammarRule* rule);
	void print_or(FILE *f, GrammarOrRule* or_rule);

	AbstractScanner* _scanner;
	GrammarNonTerminal* _all_nt;
	
	typedef struct {
	  const char *sym;
	  Ident in_nt;
	  GrammarRule* rule;
	  bool is_keyword;
	} expect_t;
	expect_t _expect[200];
	int _nr_exp_syms;
	TextFilePos _f_file_pos;
	void expected_string(TextFilePos& pos, const char *s, bool is_keyword);
	Ident _current_nt;
	GrammarRule* _current_rule;
	
	bool _debug_nt;
	bool _debug_parse;
	bool _debug_scan;
};

#endif // _INCLUDED_ABSTRACTPARSER_H

