
#include "StdAfx.h"
#include "CDFSyntaxColorer.h"

CDFSyntaxColorState::CDFSyntaxColorState()
{
	state = 0;
}

CDFSyntaxColorState::CDFSyntaxColorState(CDFSyntaxColorState* original)
{
	stateStack = original->stateStack;
	state = original->state;
}



CDFSyntaxColorer::CDFSyntaxColorer()
{
	defaultTextColor = ::GetSysColor(COLOR_WINDOWTEXT);
	defaultBkColor = ::GetSysColor(COLOR_WINDOW);
	noTextColor = defaultBkColor;
	selectTextColor = RGB(255, 255, 255);
	selectBkColor = RGB(0, 0, 0);
}


#define STATE(X) 	(X & 7)
#define ST_TOP		 0
#define ST_ERROR	 1
#define ST_GEOM		 2
#define ST_PAR		 3
#define ST_TEXT		 4
#define ST_END		 5
#define ST_IN_TEXT	 8


#define SYM_UNDEFINED		0
#define SYM_LIST_OPEN		1
#define SYM_LIST_CLOSE		2
#define SYM_REC_OPEN		3
#define SYM_REC_CLOSE		4
#define SYM_IDENT			5
#define SYM_LIST_NAME_OPEN	6
#define SYM_REC_NAME_OPEN	7
#define SYM_NUMBER			8

#define COLOR_TEXT_ERROR	RGB(128, 0, 0)
#define COLOR_BK_ERROR		RGB(255, 240, 240)
#define COLOR_TEXT_COMMENT	RGB(64, 64, 220)
#define COLOR_TEXT_STRING	RGB(0, 128, 0)
#define COLOR_BK_STRING		RGB(240, 255, 240)
#define COLOR_TEXT_GREEK	RGB(0, 64, 128)
#define COLOR_BK_GREEK		RGB(240, 245, 255)

void CDFSyntaxColorer::parseLine(const char *line, CDFSyntaxColorState &CDFstate, int &lineFlags)
{
	SharedStack<char> *stateStack = &CDFstate.stateStack;
	int state = CDFstate.state;

	if (line[0] == '\0')
		return;

	if (state == ST_ERROR)
	{	startColor(0, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
		return;
	}

	char buf[100000];

	if (line[0] == '%')
	{
		startColor(0, COLOR_TEXT_COMMENT, defaultBkColor);
		CDFstate.state = state;
		return;
	}

	int i = 0;
	if (line[0] == '\t' || line[0] == ' ')
	{
		while (line[i] == '\t' || line[i] == ' ')
			i++;
		startColor(0, defaultTextColor, defaultBkColor);
	}

	if (state & ST_IN_TEXT)
	{
		int k = 0;
		int begin = i;
		for (; line[i] != '\0' && line[i] != '}'; i++)
			if (k < 99999)
				buf[k++] = line[i];
		buf[k] = '\0';
		startColor(begin, COLOR_TEXT_STRING, COLOR_BK_STRING);
		if (line[i] == '}')
		{	endColor(i);
			state &= ~ST_IN_TEXT;
			i++;
			if (stateStack->topIs('='))
				stateStack->set(')');
		}
		/*if (STATE(dwParseState) == ST_G)
		{	int error_pos;
			startColor(begin, COLOR_TEXT_GREEK, COLOR_BK_STRING);
			if (!correct_greek(buf, error_pos))
			{	startColor(begin + error_pos + 1, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
				parseState->state = ST_ERROR;
				return;
			}
		}
		else
		{  
			..
		}*/
	}

	while (line[i] != '\0')
	{
		// scan symbol
		COLORREF fg = defaultTextColor;
		COLORREF bk = defaultBkColor;

		if (line[i] == '\t' || line[i] == ' ')
		{
			startColor(i, fg, bk);
			while (line[i] == '\t' || line[i] == ' ')
				i++;
		}

		int begin = i;
		if (line[i] == '[')
		{	i++;
			
			stateStack->push(']');
		}
		else if (line[i] == ']')
		{	i++;

			if (!stateStack->topIs(']'))
				state = ST_ERROR;
			else
			{
				stateStack->pop();
				if (stateStack->topIs('='))
					stateStack->set(')');
			}
		}
		else if (line[i] == '(')
		{	i++;

			stateStack->push(')');
		}
		else if (line[i] == ')')
		{	i++;

			if (!stateStack->topIs(')'))
				state = ST_ERROR;
			else
			{
				stateStack->pop();
				if (stateStack->topIs('='))
					stateStack->set(')');
			}
		}
		else if (isalpha(line[i]) || line[i]=='_')
		{	int k = 0;
			for (; isalnum(line[i]) || line[i]=='_'; i++)
				if (k < 99999)
					buf[k++] = line[i];
			buf[k] = '\0';
			if (line[i] == '[')
			{	i++;

				stateStack->push(']');
			}
			else if (line[i] == '(')
			{	i++;

				stateStack->push(')');
			}
			else if (line[i] == '=')
			{
				if (!stateStack->topIs(')'))
					state = ST_ERROR;
				else
				{	i++;
					stateStack->set('=');
				}
			}
			else
			{
				state = ST_ERROR;
				// not valid
			}
		}
		else if (isdigit(line[i]))
		{	
			int begin = i;
			for (i++; isdigit(line[i]); i++);
			if (stateStack->topIs('='))
				stateStack->set(')');
		}
		else if (line[i] == '{')
		{
			i++;
			begin = i;
			state = state | ST_IN_TEXT;
			
			int k = 0;
			for (; line[i] != '\0' && line[i] != '}'; i++)
				if (k < 99999)
					buf[k++] = line[i];
			buf[k] = '\0';
			if (line[i] == '}')
			{	color(begin, i, COLOR_TEXT_STRING, COLOR_BK_STRING);
				state &= ~ST_IN_TEXT;
				i++;
				if (stateStack->topIs('='))
					stateStack->set(')');
			}
			else
			{
				fg = COLOR_TEXT_STRING;
				bk = COLOR_BK_STRING;
			}

			/*if (   STATE(dwParseState) == ST_G
					|| STATE(dwParseState) == ST_P_G
					|| STATE(dwParseState) == ST_IT_G
					|| STATE(dwParseState) == ST_BOLD_G)
				{	
					fg = COLOR_TEXT_GREEK;
					bk = COLOR_BK_STRING;
					int error_pos;
					if (!correct_greek(buf, error_pos))
					{	startColor(begin, fg, bk);
						startColor(begin + error_pos + 1, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
						parseState->state = ST_ERROR;
						return;
					}
				}
				else
				{	
				}*/
		}
		else
			state = ST_ERROR;

		if (state == ST_ERROR)
		{	fg = COLOR_TEXT_ERROR;
			bk = COLOR_BK_ERROR;
		}

		startColor(begin, fg, bk);
		if (state == ST_ERROR)
			break;
	}

	CDFstate.state = state;

}

