#include "stdafx.h"
#include "ScrSyntaxColorer.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

ScrSyntaxColorState::ScrSyntaxColorState()
{
	state = 0;
}

ScrSyntaxColorState::ScrSyntaxColorState(ScrSyntaxColorState* original)
{
	stateStack = original->stateStack;
	state = original->state;
}

//	C++ keywords (MSVC5.0 + POET5.0)
static LPTSTR s_apszScrKeywordList[] =
{
	_T("false"),
	_T("true"),
	_T("bool"),
	_T("int"),
	_T("double"),
	_T("string"),
	_T("textbuffer"),
	_T("image"),
	_T("color"),
	_T("point"),
	_T("if"),
	_T("else"),
	_T("while"),
	_T("for"),
	_T("from"),
	_T("to"),
	NULL
};

static BOOL IsScrKeyword(LPCTSTR pszChars, int nLength)
{
	for (int L = 0; s_apszScrKeywordList[L] != NULL; L ++)
	{
		if (strncmp(s_apszScrKeywordList[L], pszChars, nLength) == 0
				&& s_apszScrKeywordList[L][nLength] == 0)
			return TRUE;
	}
	return FALSE;
}

static BOOL IsScrNumber(LPCTSTR pszChars, int nLength)
{
	if (nLength > 2 && pszChars[0] == '0' && pszChars[1] == 'x')
	{
		for (int I = 2; I < nLength; I++)
		{
			if (isdigit(pszChars[I]) || (pszChars[I] >= 'A' && pszChars[I] <= 'F') ||
										(pszChars[I] >= 'a' && pszChars[I] <= 'f'))
				continue;
			return FALSE;
		}
		return TRUE;
	}
	if (! isdigit(pszChars[0]))
		return FALSE;
	for (int I = 1; I < nLength; I++)
	{
		if (! isdigit(pszChars[I]) && pszChars[I] != '+' &&
			pszChars[I] != '-' && pszChars[I] != '.' && pszChars[I] != 'e' &&
			pszChars[I] != 'E')
			return FALSE;
	}
	return TRUE;
}

#define COOKIE_COMMENT			0x0001
#define COOKIE_PREPROCESSOR		0x0002
#define COOKIE_EXT_COMMENT		0x0004
#define COOKIE_STRING			0x0008
#define COOKIE_CHAR				0x0010

#define COLOR_WHITESPACE	RGB(255, 255, 255)
#define COLOR_NORMALTEXT	RGB(  0,   0,   0)
#define COLOR_PREPROCESSOR	RGB(128,   0,   0)
#define COLOR_COMMENT		RGB( 64,  64, 220)
#define COLOR_NUMBER		RGB(128,   0,   0)
#define COLOR_OPERATOR		RGB(  0,   0,   0)
#define COLOR_STRING		RGB(  0, 128,   0)
#define COLOR_BK_STRING		RGB(240, 255, 240)
#define COLOR_KEYWORD		RGB(  0,  64, 128)
#define COLOR_SELBKGND		RGB(  0,   0,   0)
#define COLOR_SELTEXT		RGB(255, 255, 255)
#define COLOR_TEXT_ERROR	RGB(128, 0, 0)
#define COLOR_BK_ERROR		RGB(255, 240, 240)

ScrSyntaxColorer::ScrSyntaxColorer()
{
	defaultTextColor = ::GetSysColor(COLOR_WINDOWTEXT);
	defaultBkColor = ::GetSysColor(COLOR_WINDOW);
	noTextColor = defaultBkColor;
	selectTextColor = RGB(255, 255, 255);
	selectBkColor = RGB(0, 0, 0);
}

void ScrSyntaxColorer::parseLine(const char* line, ScrSyntaxColorState &Scrstate, int &lineFlags)
{
	SharedStack<char> *stateStack = &Scrstate.stateStack;
	int state = Scrstate.state;

	BOOL bFirstChar     = (state & ~COOKIE_EXT_COMMENT) == 0;
	BOOL bRedefineBlock = TRUE;
	BOOL bDecIndex  = FALSE;
	int nIdentBegin = -1;
	for (int I = 0; ; I++)
	{
		if (bRedefineBlock)
		{
			int nPos = I;
			if (bDecIndex)
				nPos--;
			if (state & (COOKIE_COMMENT | COOKIE_EXT_COMMENT))
			{
				startColor(nPos, COLOR_COMMENT, COLOR_WHITESPACE);
			}
			else
			if (state & (COOKIE_CHAR | COOKIE_STRING))
			{
				startColor(nPos, COLOR_STRING, COLOR_BK_STRING);
			}
			else
			if (state & COOKIE_PREPROCESSOR)
			{
				startColor(nPos, COLOR_PREPROCESSOR, COLOR_WHITESPACE);
			}
			else
			{
				startColor(nPos, COLOR_NORMALTEXT, COLOR_WHITESPACE);
			}
			bRedefineBlock = FALSE;
			bDecIndex      = FALSE;
		}

		if (line[I] == '\0')
			break;

		if (state & COOKIE_COMMENT)
		{
			startColor(I, COLOR_COMMENT, COLOR_WHITESPACE);
			state |= COOKIE_COMMENT;
			break;
		}

		//	String constant "...."
		if (state & COOKIE_STRING)
		{
			if (line[I] == '"' && (I == 0 || line[I - 1] != '\\' || (I >= 2 && line[I - 2] == '\\')))
			{
				state &= ~COOKIE_STRING;
				bRedefineBlock = TRUE;
				endColor(I);
			}
			continue;
		}

		//	Char constant '..'
		if (state & COOKIE_CHAR)
		{
			if (line[I] == '\'' && (I == 0 || line[I - 1] != '\\' || (I >= 2 && line[I - 2] == '\\')))
			{
				state &= ~COOKIE_CHAR;
				bRedefineBlock = TRUE;
				endColor(I);
			}
			continue;
		}

		//	Extended comment /*....*/
		if (state & COOKIE_EXT_COMMENT)
		{
			if (I > 0 && line[I] == '/' && line[I - 1] == '*')
			{
				state &= ~COOKIE_EXT_COMMENT;
				bRedefineBlock = TRUE;
			}
			continue;
		}

		if (I > 0 && line[I] == '/' && line[I - 1] == '/')
		{
			startColor(I - 1, COLOR_COMMENT, COLOR_WHITESPACE);
			state |= COOKIE_COMMENT;
			break;
		}

		//	Preprocessor directive #....
		if (state & COOKIE_PREPROCESSOR)
		{
			if (I > 0 && line[I] == '*' && line[I - 1] == '/')
			{
				startColor(I - 1, COLOR_COMMENT, COLOR_WHITESPACE);
				state |= COOKIE_EXT_COMMENT;
			}
			continue;
		}

		//	Normal text
		if (line[I] == '"')
		{
			startColor(I+1, COLOR_STRING, COLOR_BK_STRING);
			state |= COOKIE_STRING;
			continue;
		}
		if (line[I] == '\'')
		{
			startColor(I+1, COLOR_STRING, COLOR_BK_STRING);
			state |= COOKIE_CHAR;
			continue;
		}
		if (I > 0 && line[I] == '*' && line[I - 1] == '/')
		{
			startColor(I - 1, COLOR_COMMENT, COLOR_WHITESPACE);
			state |= COOKIE_EXT_COMMENT;
			continue;
		}

		if (bFirstChar)
		{
			if (line[I] == '#')
			{
				startColor(I, COLOR_PREPROCESSOR, COLOR_WHITESPACE);
				state |= COOKIE_PREPROCESSOR;
				continue;
			}
			if (! isspace(line[I]))
				bFirstChar = FALSE;
		}


		if (nIdentBegin >= 0 && !(isalnum(line[I]) || line[I] == '_' || line[I] == '.'))
		{
			if (IsScrKeyword(line + nIdentBegin, I - nIdentBegin))
			{
				startColor(nIdentBegin, COLOR_KEYWORD, COLOR_WHITESPACE);
			}
			else if (IsScrNumber(line + nIdentBegin, I - nIdentBegin))
			{
				startColor(nIdentBegin, COLOR_NUMBER, COLOR_WHITESPACE);
			}
			bRedefineBlock = TRUE;
			bDecIndex = TRUE;
			nIdentBegin = -1;
		}

		if (line[I] == '(')
		{	
			stateStack->push(')');
		}
		else if (line[I] == ')')
		{	
			if (stateStack->topIs(')'))
				stateStack->pop();
			else
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
		}
		else if (line[I] == '[')
		{	
			stateStack->push(']');
		}
		else if (line[I] == ']')
		{	if (stateStack->topIs(']'))
				stateStack->pop();
			else
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
		}
		else if (line[I] == '{')
		{	
			if (stateStack->topIs(')') || stateStack->topIs(']'))
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
			stateStack->push('}');
		}
		else if (line[I] == '}')
		{	
			if (stateStack->topIs('}'))
				stateStack->pop();
			else
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
		}
		else if (isalnum(line[I]) || line[I] == '_' || line[I] == '.')
		{
			if (nIdentBegin == -1)
				nIdentBegin = I;
		}
		else
		{
		}
	}

	if (nIdentBegin >= 0)
	{
		if (IsScrKeyword(line + nIdentBegin, I - nIdentBegin))
		{
			startColor(nIdentBegin, COLOR_KEYWORD, COLOR_WHITESPACE);
		}
		else
		if (IsScrNumber(line + nIdentBegin, I - nIdentBegin))
		{
			startColor(nIdentBegin, COLOR_NUMBER, COLOR_WHITESPACE);
		}
	}

	if (I > 0 && line[I - 1] != '\\')
		state &= COOKIE_EXT_COMMENT;

	Scrstate.state = state;
}
