// gfl.h : interface to GFL
//
/////////////////////////////////////////////////////////////////////////////

#if !defined(AFX_GFL_H__INCLUDED_)
#define AFX_GFL_H__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

#include "libgfl.h"

const int m_undoBufferSize = 100;

class GFLBitmap
{
public:
	GFLBitmap() : m_gflBitmap(0), m_dib_info(0), m_undoStart(0), m_undoCurrent(0), m_undoEnd(0), m_undoSaved(-1) {}
	~GFLBitmap() { clear(); }

	static bool initialized();

	void clear();
	int width();
	int height();

	bool load(const char* filename);
	void newImage(int width, int height);
	bool save(const char* filename);

	void drawImageStretch(HDC hdc, int XDest, int YDest, int nDestWidth, int nDestHeight, int XSrc, int YSrc);

	bool getColor(int x, int y, double &r, double &g, double &b);
	bool getInterpolatedColor(double x, double y, double &r, double &g, double &b);
	bool setColor(int x, int y, int r, int g, int b);

	bool changed() { return GetModified() == TRUE; }

	static void init();
	static bool isImageFormat(LPCTSTR lpszPathName);

	//	Undo/Redo
	BOOL CanUndo();
	BOOL CanRedo();
	BOOL Undo();
	BOOL Redo();

protected:
	virtual void SetModified(BOOL bModified = TRUE) {}
	virtual BOOL GetModified() { return FALSE; }

private:
	GFL_BITMAP* m_gflBitmap;
	BITMAPINFOHEADER* m_dib_info;

	struct UndoItem {
		long x;
		long y;
		long undoColor;
		long redoColor;
	};
	
	UndoItem m_undoBuffer[m_undoBufferSize];
	long m_undoStart;
	long m_undoCurrent;
	long m_undoEnd;
	long m_undoSaved;
};

#endif // AFX_GFL_H__INCLUDED_